// This file is part of the brownmove simulation package
// 
// V1.2  Apr. 2011
// 
// (C) 2003 - 2011 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#ifndef GEOMSHAPE_H
#define GEOMSHAPE_H

#include <iostream>
#include <string>
#include <time.h>
#include <sstream>

#include "types.h"
#include "vectormath.h"
#include "constants.h"
#include "random.h"

#include "shape.h"

using namespace std;

namespace brown{

	//! Limit to the maximal step due to the interparticle and external
	//! forces in units of A0. If this limit is exceeded, the displacement 
	//! is reduced and a warning printed to STDERR. 
	//! Remedy: reduce integration timestep
	const double MAX_BROWNIAN_STEP = 8.0;  // [A0]


	/**
	 *@author jni, tg
	 */

	//! This class implements the diffusive properties of the particle, it converts 
	//! forces into displacements, GeomShape also may contain reaction spheres that
	//! trigger a reaction.

	// Initially, the code was prepared for anisotropic diffusion (by using matrices),
	// but this will have to wait until someone figures out how to handle HI between
	// non-spherical particles
	
class GeomShape : public Shape {
	public:
	//! Constructor; initializes diffusion matrices to zero
	GeomShape();
	
	//! Destructor
	~GeomShape();

	//! Output to string for diagnostics
	virtual std::string describe();
	
	//! Check whether distance criteria that may induce a reaction are met
	virtual bool checkCollision(GeomShape *other);
	
	//! restores the parts' positions from fullPosition
	virtual void recalculate(const fullPosition &p);

	//! @name Set/get diffusion constants, gamma and exponential corrections for displacements and Brownian steps (includes timestep), 
	//! @{
	//
	//! Set D0_trans for displacement from forces and calcualte related helper variables
	void setDTrans(const double &d0_trans, const double &timestep, const double &beta, const double &m, bool verbose);
	
	//! set rotational diffusion constant and calculate related helper variables
	void setDRot(const double &d0_rot, const double &timestep, const double &beta, const double &m, const double &r, bool verbose);
	
	//! Return typical Brownian step for translation
	const double getA0Trans() {return(A0_Trans);} 	
	//! Return typical Brownian step for rotation
	const double getA0Rot() {return(A0_Rot);}

	//! Return force for a typical translational Brownian step
	const double getF0Trans() {return(F0_Trans);}
	//! Return torque for a typical rotational Brownian step
	const double getF0Rot()   {return(F0_Rot);}

	double* getRandom() {return(Random_Storage);}
	//! @}


	//! set the flag whether the molecule does rotate or not (default: no rot = false)
	inline void setDoesRotate(bool flag) { Does_Rotate = flag; }
	//! retrun whether the gestalt does rotate
	bool getDoesRotate() { return(Does_Rotate); }	
	//! set the flag whether the molecule is spherical or not (default: spherical = true)
	inline void setisSpherical(bool flag) { Is_Spherical = flag; }

	//! needed to unwind the generation of random forces into an own subroutine
	void generate_new_random_forces();
	
	//! initialize random forces of the Gestalt
	void initRandomForces(fullForce &randomForce);
	
	//! Calculate and apply corrections for finite damping (Langevin propagation)
	//! @param[in,out] externalForce Pointer to the sum of the external forces acting on the gestalt
	//! @param[in,out] randomForce Random forces acting on gestalt
	void correctDamping(fullForce &externalForce, fullForce &randomForce);
	
	//! Calculate the displacement during one tinestep from the sum of forces and torques
	//! @param[out] dfp the change in fullposition during one timestep
	//! @param[in] f actual forces
	//! @param[in] t actual torque
	void getDisplacement(fullPosition &dfp, const force &f, const torque &t);

	//! Adds another sphere with position and radius to the vector.
	//! Also calculates the relative position of the sphere from the given molecule center.
	void addSphere(geomPos gp, const position &moleculeCenter);

	//! @name Hooks for structure dumping
	// @{
	const double getD0Trans() { return(myD0_trans); }
	const double getD0Rot()   { return(myD0_rot); }
	const double getMass()    { return(myMass); }
	const double getRadius()  { return(myRadius); }
	// @}
	
	private:
	//! @name diffusion_coefficients : the actual values of D
	//! @{
	//
	//! Is the particle spherical or do we need a diffusion matrix for anisotropic behavior? (see above)
	bool Is_Spherical;
	//! Do we need rotation? (no need for rotation if molecule is spherically symmetric - charged sphere, e.g.)
	bool Does_Rotate;
	//! diffusion coefficients for a spherical particle. 
	double myD0_trans, myD0_rot, myMass, myRadius;
	double T_Over_Gamma_Trans;
	double T_Over_Gamma_Rot;
	double A0_Trans;	// a0 = sqrt(4 D tau)
	double A0_Rot;
	double F0_Trans;	// f0 = sqrt(4 (kT)^2 tau / D)
	double F0_Rot;

	bool applyCorrections;
	double Exp_Coeff_trans, prefactorF_trans;
	double Exp_Coeff_rot, prefactorF_rot;
	double prefactorForce, prefactorV0_trans;
	double prefactorTorque, prefactorV0_rot;
	doublevec v0_trans_ext, v0_trans_rnd;
	doublevec v0_rot_ext, v0_rot_rnd;
	//! @}

	//! storage for DUALPASS integrators.
	double Random_Storage [6];

	//! vectors of the spheres used for collision checking
	vector<geomPos> Spheres;
	vector<geomPos> Actual_Spheres;
};
}

#endif
