// This file is part of the brownmove simulation package
// 
// V1.2  Apr. 2011
// 
// (C) 2003 - 2011 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include <iostream>
#include <string>
#include <sstream>

#include "bondshape.h"

namespace brown{
	
	BondShape::BondShape(){
		Params_Vector = NULL;
	}
	
	BondShape::BondShape(const BondShape &other){
		Hooks = other.Hooks;
		Actual_Hooks = other.Actual_Hooks;
		Params_Vector = other.Params_Vector;
	}
	

	BondShape::~BondShape(){
		Hooks.clear();
		Actual_Hooks.clear();
	}
	
	
	std::string BondShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual positions";
		vector<bondPos>::iterator mlc;
		for(mlc = Actual_Hooks.begin(); mlc != Actual_Hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		ausgabe <<" :";
		for(mlc = Hooks.begin(); mlc != Hooks.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		return(ausgabe.str());
	}
	
	
	
	void BondShape::recalculate(const fullPosition &p) 
	{
		vector<bondPos>::iterator act = Actual_Hooks.begin();
		for(vector<bondPos>::iterator mlc = Hooks.begin(); mlc != Hooks.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			doublevec_set((*act).dr, (*act).pos); //slightly optimized
			positionMove((*act).pos, p.pos);
		}
	}


	
	void BondShape::addHook(bondPos bp, const position &moleculeCenter){
		doublevec_set(bp.dr, bp.pos);
		doublevec_sub(bp.dr, moleculeCenter);
		Hooks.push_back(bp);
		Actual_Hooks.push_back(bp);
	}
	
	
	bool BondShape::calculateImpact(fullForce &myForce, fullForce &otherForce, BondShape *bs)
	{
		long int idx1, idx2;
		bool is_bonded = false;

		// cout <<"*** BondShape::calculateImpact:   myForce.f =    "<< myForce.f[0] << " " << myForce.f[1] << " " << myForce.f[2] << endl;
		// cout <<"*** BondShape::calculateImpact:   otherForce.f = "<< otherForce.f[0] << " " << otherForce.f[1] << " " << otherForce.f[2] << endl;
		
		// first interact all hooks with each other -- nothing happens, if there are none
		// Therefore, begin with the other's hooks (could be a wall..) -- we have at least one hooks, or we would not be here.
		for(vector<bondPos>::iterator mlc2 = bs->Actual_Hooks.begin(); mlc2 != bs->Actual_Hooks.end(); ++mlc2)
		{
			idx2 = (*mlc2).idx;
			
			// cout <<"*** BondShape::calculateImpact  mlc2:   "<<"  " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;
			
			for(vector<bondPos>::iterator mlc1 = Actual_Hooks.begin(); mlc1 != Actual_Hooks.end(); ++mlc1)
			{
				idx1 = (*mlc1).idx;
				// cout <<"*** BondShape::calculateImpact  mlc1: "<< "   " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;

				// 1. check whether the two indices are idential 
				if(idx1 == idx2)
				{
					position dr;
					double prefactor;
					double r12, dist;
					force df;
					torque dt;
					
					is_bonded = true;
					
					// calculate relative distances etc
					doublevec_set(dr, (*mlc2).pos);
					doublevec_sub(dr, (*mlc1).pos);
					r12 = doublevec_length(dr);
					dist = r12 - (*Params_Vector)[idx1].R_0;

					// --- here comes the interaction itself
					prefactor = ((*Params_Vector)[idx1].k2 + (*Params_Vector)[idx1].k4 * dist * dist) * dist;
					
					doublevec_set(df, dr);					// for the direction
					doublevec_mult(df, (prefactor/r12));	// normalize dr by its length
					
					// add force to previous results
					doublevec_add(myForce.f, df);
					doublevec_sub(otherForce.f, df);
					
					// calculate and add torques
					doublevec_kreuzprod(dt, (*mlc1).dr, df);
					doublevec_add(myForce.t, dt);
					doublevec_kreuzprod(dt, (*mlc2).dr, df);
					doublevec_sub(otherForce.t, dt);
				}
// 				else // do nothing (except in debug mode)
// 				{
// 					cout << "BondShape::calculateImpact: indices different"<< endl;
// 				}
				
				
 				// cout <<"                 df = " << df[0] << " " << df[1] << " " << df[2] << "   prefactor = " << prefactor << endl;
 				// cout <<"                 dt = " << dt[0] << " " << dt[1] << " " << dt[2] << endl;
			}
			
// 			cout <<"*** BondShape::calculateImpact:   myForce.f =    "<< myForce.f[0] << " " << myForce.f[1] << " " << myForce.f[2] << endl;
// 			cout <<"                                   myForce.t =    "<< myForce.t[0] << " " << myForce.t[1] << " " << myForce.t[2] << endl;
// 			cout <<"*** BondShape::calculateImpact:   otherForce.f = "<< otherForce.f[0] << " " << otherForce.f[1] << " " << otherForce.f[2] << endl;
// 			cout <<"                                   otherForce.t = "<< otherForce.t[0] << " " << otherForce.t[1] << " " << otherForce.t[2] << endl;
		}
		return(is_bonded);
	}

	
	int BondShape::size() {
		return(Hooks.size());
	}
	
	
	void BondShape::dumpStructure() {
		std::cout << "\n\t\tBondShape myBonds" << std::endl;
		
		for(std::vector<bondPos>::iterator it = Hooks.begin(); it != Hooks.end(); it++) {
			std::cout << "\t\t\t"<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<" "<< (*Params_Vector)[(*it).idx].label << std::endl;
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	void BondShape::listHooks(std::vector<bondPos> &haken) 
	{
		for(std::vector<bondPos>::iterator it = Actual_Hooks.begin(); it != Actual_Hooks.end(); it++) {
			haken.push_back((*it));
		}
	}
}
