// This file is part of the brownmove simulation package
// 
// V1.2  Apr. 2011
// 
// (C) 2003 - 2011 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "estatshape.h"

namespace brown{
	// access to global variables
	extern double KAPPA;
	extern double EPS_H2O;

	EstatShape::EstatShape(){
		interAction.interaction = NULL;
	}
	
	EstatShape::EstatShape(const EstatShape &other){
		charges = other.charges;
		actual_charges = other.actual_charges;
		interAction = other.interAction;
	}
	
	EstatShape::~EstatShape(){
		charges.clear();
		actual_charges.clear();
	}
	
	
	std::string EstatShape::describe(){
		ostringstream ausgabe("", ios::out);
		int i=0;
		
		ausgabe << "actual charge positions";
		vector<chargePos>::iterator mlc;
		for(mlc = actual_charges.begin(); mlc != actual_charges.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		ausgabe <<" :";
		for(mlc = charges.begin(); mlc != charges.end(); ++mlc, i++){
			ausgabe << " : "<<(*mlc).pos[0]<<" "<<(*mlc).pos[1]<<" "<<(*mlc).pos[2] ; 
		}
		return(ausgabe.str());
	}
	
	
	void EstatShape::recalculate(const fullPosition &p) 
	{
		vector<chargePos>::iterator act = actual_charges.begin();
		for(vector<chargePos>::iterator mlc = charges.begin(); mlc != charges.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
		}
	}
	
	
	void EstatShape::addCharge(chargePos cp, const position &moleculeCenter){
		doublevec_set(cp.dr, cp.pos);
		doublevec_sub(cp.dr, moleculeCenter);
		charges.push_back(cp);
		actual_charges.push_back(cp);
	}
	
	
	void EstatShape::calculateImpact(fullForce &myForce, fullForce &otherForce, EstatShape *es, const bool isPeriodic[3], const doublevec &cellSize)
	{
		position dr, ri, dr_temp;
		double r12, prefactor, B12, help, Rmin;
		force df;
		torque dt;
		functionPointer otherFP;
		
		// first interact all point charges with each other -- nothing happens, if there are no charges
		// Begin with the other's charges -- if it is a wall it may have none
		for(vector<chargePos>::iterator mlc2 = es->actual_charges.begin(); mlc2 != es->actual_charges.end(); ++mlc2)
		{
			// cout <<"*** EstatShape::calculateImpact  mlc2:   "<<"  " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;

			// Now my own charges (we wouldn't be here if there were none)
			for(vector<chargePos>::iterator mlc1 = actual_charges.begin(); mlc1 != actual_charges.end(); ++mlc1)
			{
				// cout <<"*** EstatShape::calculateImpact  mlc1: "<< "   " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;
				
				// calculate relative distances etc
				doublevec_set(dr, (*mlc2).pos);
				doublevec_sub(dr, (*mlc1).pos);
				B12 = (*mlc1).r + (*mlc2).r;
				
				// this is the distance where shielded Coulomb and plain Coulomb have the same interaction energy
				if(KAPPA > 0.0) {
					Rmin = B12 - log(1.0 + KAPPA*B12) / KAPPA;
				} else {
					Rmin = 0.0;
				}
				
				
				double dx = 0.0;
				double dy = 0.0;
				double dz = 0.0;
				int ix = 2, iy = 2, iz = 2;
				
				// if periodic, start with displaced charges
				if(isPeriodic[0]) {
					dx = -cellSize[0];
					ix = 0;
				}
				if(isPeriodic[1]) {
					dy = -cellSize[1];
					iy = 0;
				}
				if(isPeriodic[2]) {
					dz = -cellSize[2];
					iz = 0;
				}
				
				while(true)
				{
					// this is for the (quasi) periodic setup
					dr_temp[0] = dr[0] + dx;
					dr_temp[1] = dr[1] + dy;
					dr_temp[2] = dr[2] + dz;
					
					r12 = doublevec_length(dr_temp);
					
					// // Old version:
					// // check if distance between charges is at least as large as the sum of their insertion depths
					// if(r12 <= B12) {
					// 	r12 = B12;
					// }
					// New version: linearized where shielded and plain Coulomb potential have same energy
					if(r12 <= Rmin) {
						r12 = Rmin;
					}
					
					
					// this is the shielded Coulomb force itself
					prefactor = (*mlc1).ch * (*mlc2).ch /(4.0*M_PI * EPS_0 * EPS_H2O *(1.0 + KAPPA*B12));
					help = exp(-KAPPA * (r12 - B12)) * (KAPPA + 1.0/r12) / (r12*r12);
					prefactor *= help;
					doublevec_set(df, dr_temp);				// for the direction
					doublevec_mult(df, prefactor);
					
					// add force to previous results
					doublevec_add(otherForce.f, df);
					doublevec_sub(myForce.f, df);
					
					// calculate and add torques
					doublevec_kreuzprod(dt, (*mlc2).dr, df);
					doublevec_add(otherForce.t, dt);
					doublevec_kreuzprod(dt, (*mlc1).dr, df);
					doublevec_sub(myForce.t, dt);


					// for the quasi periodic boundaries: add displaced charges and redo above evaluation
					if(ix < 2) {
						ix++;
						dx += cellSize[0];
					}
					else
					{
						if(isPeriodic[0]) {
							ix = 0;
							dx = -cellSize[0];
						}
						
						if(iy < 2) {
							iy++;
							dy += cellSize[1];
						}
						else
						{
							if(isPeriodic[1]) {
								iy = 0;
								dy = -cellSize[1];
							}
												
							if(iz < 2) {
								iz++;
								dz += cellSize[2];
							}
							else
							{
								break;
							}
						}
					}
				}
			}
		}
		
		
		// now -- for walls -- check, if there is some analytic interaction
		// i.e., the pointer is not NULL anymore
		otherFP = es->getInteraction();
		if(otherFP.interaction) 
		{
			// cout <<"*** EstatShape::calculateImpact: calculating analytical Coulomb interaction" << endl;
			
			position *wallPos;
			direction *wallNormal;
			position abstand;
			double dn, ergForce;
			
			// cout <<"*** EstatShape::calculateImpact: position = " << wallPos << "   Normal = " << wallNormal << endl;
			wallPos = es->getPosition();
			wallNormal = es->getNormal();
			
			// cout <<"*** EstatShape::calculateImpact: position = " << wallPos << "   Normal = " << wallNormal << endl;
			// cout <<"*** EstatShape::calculateImpact: position = " <<(*wallPos)[0]<<"  "<<(*wallPos)[1]<<"  "<<(*wallPos)[2]<< endl;
			// cout <<"*** EstatShape::calculateImpact: normal   = " <<(*wallNormal)[0]<<"  "<<(*wallNormal)[1]<<"  "<<(*wallNormal)[2]<< endl;
			
			int i = 1;
			for(vector<chargePos>::iterator mlc2 = actual_charges.begin(); mlc2 != actual_charges.end(); ++mlc2, ++i)
			{
				// cout <<"*** EstatShape::calculateImpact: charge " << i << "  " << (*mlc2).ch << endl;
				
				doublevec_set(abstand, (*mlc2).pos);
				doublevec_sub(abstand, (*wallPos));
				dn = doublevec_scalprod(abstand, (*wallNormal));
				// 	cout << "*** EstatShape::calculateImpact: abstand =     " << abstand[0] << " "<< abstand[1] << " "<< abstand[2] << " -> " << dn <<endl;
				
				if(dn <= 1.0e-3) {
					// cout << "===== EstatShape::calculateImpact: dn = " << dn << " =====" << endl;
					dn = 1.0e-3;
				}
				
				doublevec_set(df, (*wallNormal));
				ergForce = otherFP.interaction(dn, (*mlc2).ch);
				doublevec_mult(df, ergForce);
				
				// cout << "*** EstatShape::calculateImpact: df =           " << df[0] << " "<< df[1] << " "<< df[2] << endl;
				// cout << "*** EstatShape::calculateImpact: myForce.f = " << myForce.f[0] << " "<< myForce.f[1] << " "<< myForce.f[2] << endl;
				
				doublevec_add(myForce.f, df);
				doublevec_kreuzprod(dt, (*mlc2).dr, df);
				doublevec_add(myForce.t, dt);
				
				// cout << "*** EstatShape::calculateImpact: myForce.f = " << myForce.f[0] << " "<< myForce.f[1] << " "<< myForce.f[2] << endl;
			}
		}
	}

	
	
	void EstatShape::dumpStructure() {
		std::cout << "\n\t\tEstatShape myCharges" << std::endl;
		
		for(std::vector<chargePos>::iterator it = charges.begin(); it != charges.end(); it++) {
			std::cout << "\t\t\t"<< (*it).pos[0]<<" "<<(*it).pos[1]<<" "<<(*it).pos[2]<<"   "<<(*it).ch<<"   " << (*it).r<< std::endl;
		}
		
		std::cout << "\t\tEnd" << std::endl;
	}

	
}
