// This file is part of the brownmove simulation package
// 
// V1.2  Apr. 2011
// 
// (C) 2003 - 2011 Tihamer Geyer
// tihamer.geyer@bioinformatik.uni-saarland.de
// 
// Check for Updates at the Project Homepage 
// http://service.bioinformatik.uni-saarland.de/brownmove



#include "geomshape.h"

namespace brown{
	
	GeomShape::GeomShape(){
		// set diffusion coefficients to zero - molecule can't move until set to a reasonable value
		Is_Spherical = true;
		Does_Rotate = false;
		setDTrans(0.0, 0.0, 1.0, 0.0, false);		// also: no mass as a default
		setDRot(0.0, 0.0, 1.0, 0.0, 1.0, false);
		
		applyCorrections = false;
		doublevec_set(v0_trans_ext, 0.0);
		doublevec_set(v0_trans_rnd, 0.0);
		doublevec_set(v0_rot_ext, 0.0);
		doublevec_set(v0_rot_rnd, 0.0);
	}
	
	GeomShape::~GeomShape(){
		Spheres.clear();
		Actual_Spheres.clear();
	}
	

	std::string GeomShape::describe(){
		ostringstream ausgabe("", ios::out);
		
		ausgabe << " D_trans = " <<myD0_trans<<"  D_rot = "<< myD0_rot << "  mass = " << myMass << "  r = " << myRadius;
		return(ausgabe.str());
	}
	
	
	void GeomShape::setDTrans(const double &d0_trans, const double &timestep, const double &beta, const double &m, bool verbose) 
	{
		// save values
		myD0_trans = d0_trans;
		myMass = m;
		
		// for standard-BD
		T_Over_Gamma_Trans = timestep * (beta * d0_trans); 
		A0_Trans = sqrt(2.0 * d0_trans * timestep);
		// beware: undefined random forces for D0_trans = 0
		// => set to 0 because D0 = 0 impies that particle should not move at all
		if(d0_trans > 0.0) {
			F0_Trans = sqrt(2.0 / (beta*beta * d0_trans * timestep));
		} else {
			F0_Trans = 0.0;
			myMass = 0.0;		// if the particle is immobile then its mass is irrelevant (but would lead to problems further down)
		}
		
		
		// for Langevin correction
		if(myMass > 0.0)
		{
			applyCorrections = true;
			double Gamma_Trans = 1.0/(d0_trans * beta);
			double M_Over_Gamma_Trans = m/(KJOULEMOL * Gamma_Trans);
			double M_Over_GammaT_Trans = M_Over_Gamma_Trans / timestep;
			double M_Over_T_Trans = m/(KJOULEMOL * timestep);
			
			Exp_Coeff_trans = exp(-(timestep / M_Over_Gamma_Trans) );
			prefactorF_trans = (1.0 - Exp_Coeff_trans) / Gamma_Trans;
			prefactorForce = 1.0 - M_Over_GammaT_Trans * (1.0 - Exp_Coeff_trans);
			prefactorV0_trans = M_Over_T_Trans * (1.0 - Exp_Coeff_trans);
			
			if(verbose) {
				std::cerr << "Geomshape:setDTrans: A0 = " << A0_Trans << "  tau_rel_trans = " << M_Over_Gamma_Trans << "   T_over_Gamma = " << T_Over_Gamma_Trans << "   exp damping = " << Exp_Coeff_trans << "   prefactorForce = " << prefactorForce << "  prefactorV0 = " << prefactorV0_trans << std::endl;
			}
		}
		else
		{
			applyCorrections = false;
			if(verbose) {
				std::cerr << "Geomshape:setDTrans: A0 = " << A0_Trans << "  T_over_Gamma = " << T_Over_Gamma_Trans << ",   no Langevin corrections" << std::endl;
			}
		}
	}

	void GeomShape::setDRot(const double &d0_rot, const double &timestep, const double &beta, const double &m, const double &r, bool verbose) 
	{
		// save values
		myD0_rot = d0_rot;
		myRadius = r;
		myMass = m;
		
		// for std BD
		T_Over_Gamma_Rot = timestep * beta * d0_rot;
		A0_Rot = sqrt(2.0 * d0_rot * timestep);
		if(d0_rot > 0.0) {
			F0_Rot = sqrt(2.0 / (beta*beta * d0_rot * timestep));
		} else {
			F0_Rot = 0.0;
			myMass = 0.0;		// if the particle is immobile then its mass is irrelevant (but would lead to problems further down)
		}

		Does_Rotate = true;		// just for completeness, it doesn't hurt

		// for the Langevin corrections
		if(myMass > 0.0)
		{
			applyCorrections = true;
			double Gamma_Rot = 1.0/ (d0_rot * beta);
			double M_Over_Gamma_Rot = (2.0 * m * r * r/ 5.0) / (KJOULEMOL * Gamma_Rot);
			double M_Over_GammaT_Rot = M_Over_Gamma_Rot / timestep;
			double M_Over_T_Rot = (2.0 * m * r * r/ 5.0) / (KJOULEMOL * timestep);
			
			Exp_Coeff_rot = exp(- ( timestep / M_Over_Gamma_Rot) );
			prefactorF_rot = (1.0 - Exp_Coeff_rot) / Gamma_Rot;
			prefactorTorque = 1.0 - M_Over_GammaT_Rot * (1.0 - Exp_Coeff_rot);
			prefactorV0_rot = M_Over_T_Rot * (1.0 - Exp_Coeff_rot);
			
			if(verbose) {
				std::cerr << "Geomshape:setDRot: A0 = " << A0_Rot << "  tau_rel_rot = " << M_Over_Gamma_Rot << "  T_over_Gamma = " << T_Over_Gamma_Rot << "   exp damping = " << Exp_Coeff_rot << "    prefactorTorque = " << prefactorTorque << "  prefactorV0 = " << prefactorV0_rot << std::endl;
			}
		}
		else
		{
			applyCorrections = false;
			if(verbose) {
				std::cerr << "Geomshape:setDRot: A0 = " << A0_Rot << "  T_over_Gamma = " << T_Over_Gamma_Rot << ",   no Langevin corrections" << std::endl;
			}
		}
		
	}	
	
	
	void GeomShape::correctDamping(fullForce &externalForce, fullForce &randomForce)
	{
		// only correct if damping is finite
		if(applyCorrections)
		{
			// std::cerr << "GeomShape::correctDamping" << std::endl;
			
			doublevec tmp1, tmp2;
			
			// external forces, translation
			doublevec_set(tmp1, externalForce.f);
			doublevec_mult(tmp1, prefactorForce);
			doublevec_set(tmp2, v0_trans_ext);				// v at the end of the previous step
			doublevec_mult(tmp2, prefactorV0_trans);
			
			doublevec_set(externalForce.f, tmp1);
			doublevec_add(externalForce.f, tmp2);
			
			// get new v0
			doublevec_set(tmp1, externalForce.f);
			doublevec_mult(tmp1, prefactorF_trans);
			doublevec_mult(v0_trans_ext, Exp_Coeff_trans);	// remainder from previous step
			doublevec_add(v0_trans_ext, tmp1);				// plus new contribution from the current step
			
			// random forces, translation
			doublevec_set(tmp1, randomForce.f);
			doublevec_mult(tmp1, prefactorForce);
			doublevec_set(tmp2, v0_trans_rnd);
			doublevec_mult(tmp2, prefactorV0_trans);
			
			doublevec_set(randomForce.f, tmp1);
			doublevec_add(randomForce.f, tmp2);
			
			// get new v0
			doublevec_set(tmp1, randomForce.f);
			doublevec_mult(tmp1, prefactorF_trans);
			doublevec_mult(v0_trans_rnd, Exp_Coeff_trans);
			doublevec_add(v0_trans_rnd, tmp1);
						   
			// corrections for rotation only if it rotates :-)
			if(Does_Rotate)
			{
				// external forces, rotation
				doublevec_set(tmp1, externalForce.t);
				doublevec_mult(tmp1, prefactorTorque);
				doublevec_set(tmp2, v0_rot_ext);
				doublevec_mult(tmp2, prefactorV0_rot);
				
				doublevec_set(externalForce.t, tmp1);
				doublevec_add(externalForce.t, tmp2);
				
				// get new v0
				doublevec_set(tmp1, externalForce.t);
				doublevec_mult(tmp1, prefactorF_rot);
				doublevec_mult(v0_rot_ext, Exp_Coeff_rot);
				doublevec_add(v0_rot_ext, tmp1);
				
				// random forces, rotation
				doublevec_set(tmp1, randomForce.t);
				doublevec_mult(tmp1, prefactorTorque);
				doublevec_set(tmp2, v0_rot_rnd);
				doublevec_mult(tmp2, prefactorV0_rot);
				
				doublevec_set(randomForce.t, tmp1);
				doublevec_add(randomForce.t, tmp2);
				
				// get new v0
				doublevec_set(tmp1, randomForce.t);
				doublevec_mult(tmp1, prefactorF_rot);
				doublevec_mult(v0_rot_rnd, Exp_Coeff_rot);
				doublevec_add(v0_rot_rnd, tmp1);
			}
		}
	}
	 

	
	
	void GeomShape::recalculate(const fullPosition &p) 
	{
		vector<geomPos>::iterator act = Actual_Spheres.begin();
		for(vector<geomPos>::iterator mlc = Spheres.begin(); mlc != Spheres.end(); ++mlc, ++act)
		{
			doublevec_set((*act).pos, (*mlc).pos);
			doublevec_rotate((*act).pos, p.rot);
			positionMove((*act).pos, p.pos);
			doublevec_set((*act).dr, (*act).pos);
			doublevec_sub((*act).dr, p.pos);
		}
	}
	
	
	
	void GeomShape::initRandomForces(fullForce &randomForce)
	{
		for(int i = 0; i<3; ++i) {
			randomForce.f[i] = gaussran() * F0_Trans;
		}
		if(Does_Rotate) {
			for(int i = 0; i<3; ++i) {
				randomForce.t[i] = gaussran() * F0_Rot;
			}
		}
	}
	
	

	 
	
	// get fullPosition displacement from force and torque
	// includes check for simplifications: isSpherical -> only coeffs, not matrices / doesRotate -> not only translation
	// take care: includes timestep via t_over_gamma, D_rot
	void GeomShape::getDisplacement(fullPosition &dfp, const force &f, const torque &t)
	{
		// std::cerr << "GeomShape::getDisplacement:  force = " << f[0] << ", " << f[1] << ", " << f[2] << "\n";
		direction phi;
	
		if(Is_Spherical)   // only scalar coefficients
		{
			// position part
			doublevec_set(dfp.pos, f);
			doublevec_mult(dfp.pos, T_Over_Gamma_Trans);

			// rotation part
			if(Does_Rotate)
			{
				doublevec_set(phi, t);
				doublevec_mult(phi, T_Over_Gamma_Rot);

				direction2rotation(dfp.rot, phi);
			} else {
				rotation_set(dfp.rot, 0.0, 1.0, 0.0, 0.0);
			}
			
		} else {		// arbitrarily shaped molecule
			std::cerr << "GeomShape::getDisplacement: non-spherical particles are not yet implemented..." << std::endl;
			exit(-1);
		}
		
		// check if step is small enough
		double dr2 = dfp.pos[0]*dfp.pos[0] + dfp.pos[1]*dfp.pos[1] + dfp.pos[2]*dfp.pos[2];
		double ratio = dr2 / (A0_Trans*A0_Trans * MAX_BROWNIAN_STEP*MAX_BROWNIAN_STEP);
		
		// cerr << "### GeomShape::get_BrownianStep:   dfp.pos = "<<dfp.pos[0]<<"  "<<dfp.pos[1]<<"  "<<dfp.pos[2]<<endl;

		if(ratio > 1.0)
		{
			cerr << "### GeomShape::get_BrownianStep: Step too long: "<<sqrt(dr2)<<" > "<<A0_Trans*MAX_BROWNIAN_STEP<<"  dfp.pos = "<<dfp.pos[0]<<"  "<<dfp.pos[1]<<"  "<<dfp.pos[2]<<endl;
			cerr << "    GeomShape::get_BrownianStep:   ratio = " << ratio << std::endl;
			ratio = sqrt(ratio);
			dfp.pos[0] /= ratio;
			dfp.pos[1] /= ratio;
			dfp.pos[2] /= ratio;
		}
		
	}
	

	
	void GeomShape::addSphere(geomPos gp, const position &moleculeCenter){
		doublevec_set(gp.dr, gp.pos);
		doublevec_sub(gp.dr, moleculeCenter);
		Spheres.push_back(gp);
		Actual_Spheres.push_back(gp);
	}

	
	bool GeomShape::checkCollision(GeomShape *gs) {
		// cout << "*** GeomShape::checkCollision" << endl;
		
		position dr;
		double r12, B12;
		
		for(vector<geomPos>::iterator mlc1 = Actual_Spheres.begin(); mlc1 != Actual_Spheres.end(); ++mlc1)
		{
			// cout <<"*** geomShape::checkCollision  mlc1: "<< "   " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;
			
			for(vector<geomPos>::iterator mlc2 = gs->Actual_Spheres.begin(); mlc2 != gs->Actual_Spheres.end(); ++mlc2)
			{
				// cout <<"*** geomShape::checkCollision  mlc2: "<< "   " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;

				// calculate relative distances etc
				doublevec_set(dr, (*mlc2).pos);
				doublevec_sub(dr, (*mlc1).pos);
				B12 = (*mlc1).r + (*mlc2).r;
				r12 = doublevec_length(dr);
				
				// now check for collision
				if(r12 < B12) {
					// cout <<"*** geomShape::checkCollision  mlc1: "<< "   " << (*mlc1).pos[0] << "  " << (*mlc1).pos[1] << "  " << (*mlc1).pos[2] << endl;
					// cout <<"*** geomShape::checkCollision  mlc2: "<< "   " << (*mlc2).pos[0] << "  " << (*mlc2).pos[1] << "  " << (*mlc2).pos[2] << endl;
					// cout <<"*** geomShape::checkCollision: r12 = " << r12 << "  B12 = " << B12 << endl;
					return(true);
				}
			}
		}

		return false;
	}
}
